/* winops.c
 * Copyright (C) 2000, 2001, 2002 Hidetoshi Ohtomo. */

#include <stdio.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/keysym.h>

#include "heliwm.h"
#include "config.h"

static void update_indicator(int operation, int x, int y)
{
	char s[20];
#ifdef NLS
	XRectangle inkbox;
#endif

	sprintf(s, ((operation == OP_MOVE) ? "%d, %d" : "%d x %d"), x, y);
	XClearWindow(d, indicator);
#ifdef NLS
	XmbTextExtents(fontset, s, strlen(s), NULL, &inkbox);
	XmbDrawString(d, indicator, fontset, text_gc, (ind_w - inkbox.width) / 2
		- inkbox.x, (title_h - inkbox.height) / 2 - inkbox.y, s,
		strlen(s));
#else
	XDrawString(d, indicator, text_gc, (ind_w - XTextWidth(font_info, s,
		strlen(s))) / 2, font_info->ascent + V_OFFSET, s, strlen(s));
#endif
}

static void draw_band(int x, int y, int width, int height)
{
	XDrawRectangle(d, rootwin, band_gc, x, y, width - 1, height - 1);
	XDrawLine(d, rootwin, band_gc, x + width / 2, y, x + width / 2, y +
		height);
	XDrawLine(d, rootwin, band_gc, x, y + height / 2, x + width, y + height
		/ 2);
}

#ifdef FRAME
static void move(struct heliwm_window *hw)
{
	XConfigureEvent e;

	XMoveWindow(d, hw->frame, hw->x, hw->y);
#ifdef SHAPE
	if (hw->flags & SHAPED)
		XMoveWindow(d, hw->child, hw->x + frame_w, hw->y + frame_w * 2 +
			title_h);
	else
#endif
	{	/* Notify child about the move. */
		e.type = ConfigureNotify;
		e.event = hw->child;
		e.window = hw->child;
		e.x = hw->x + frame_w;
		e.y = hw->y + frame_w + title_h;
		e.width = hw->child_w;
		e.height = hw->child_h;
		e.border_width = hw->child_bw;
		e.above = None;
		e.override_redirect = False;
		XSendEvent(d, hw->child, False, StructureNotifyMask, (XEvent *)
			&e);
	}
}
#else
#define move(hw) XMoveWindow(d, hw->child, hw->x, hw->y);
#endif /* FRAME */

/* adjust() adjusts size of a window according to its parameters.
 *	*w: desired width;
 *	*h: desired height;
 *	*sw: step-sized width is returned when non-NULL;
 *	*sh: step-sized height is returned when non-NULL;
 *	flag: bitwise-or combination of ADJUST_WIDTH and ADJUST_HEIGHT. */

#define ADJUST_WIDTH 1
#define ADJUST_HEIGHT 2
static void adjust(struct heliwm_window *hw, int *w, int *h, int *sw, int *sh,
	int flag)
{
	int i;

	if ((flag & ADJUST_WIDTH) && w) {
		if (*w <= hw->base_w) {
			if ((*w = hw->base_w) == 0)
				(*w)++;
		} else if ((hw->flags & MAX_SIZE_DEFINED) && *w > hw->max_w)
			*w = hw->max_w;
		i = (*w - hw->base_w) / hw->inc_w;
		*w = hw->base_w + hw->inc_w * i;
		if (sw != NULL)
			*sw = i;
	}

	if ((flag & ADJUST_HEIGHT) && h) {
		if (*h <= hw->base_h) {
			if ((*h = hw->base_h) == 0)
				(*h)++;
		} else if ((hw->flags & MAX_SIZE_DEFINED) && *h > hw->max_h)
			*h = hw->max_h;
		i = (*h - hw->base_h) / hw->inc_h;
		*h = hw->base_h + hw->inc_h * i;
		if (sh != NULL)
			*sh = i;
	}
}

void move_or_resize(struct heliwm_window *hw, int operation, int child_or_icon)
{
	XEvent e;
	Bool done = False, changed_x = False;
	int ow, oh, px, py, i, j, x, y, width, height, a, b, dir,
		scale = DEF_SCALE_NORM;
	unsigned int state;

	if (XGrabPointer(d, rootwin, False, ButtonPressMask, GrabModeAsync,
		GrabModeAsync, None, fleur, CurrentTime) != GrabSuccess)
		return;
	if (XGrabKeyboard(d, rootwin, True, GrabModeAsync, GrabModeAsync,
		CurrentTime) != GrabSuccess)
		goto ungrab_pointer_and_return;

	if (child_or_icon == ICON) {
		x = hw->icon_x;
		y = hw->icon_y;
		width = hw->icon_w + hw->icon_bw * 2;
		height = hw->icon_h + hw->icon_bw * 2;
	} else {
		x = hw->x, y = hw->y;
#ifdef FRAME
		width = hw->w;
		height = hw->h + ((hw->flags & SHAPED) ? hw->child_h +
			hw->child_bw * 2 : 0);
#else
		width = hw->child_w + hw->child_bw * 2;
		height = hw->child_h + hw->child_bw * 2;
#endif
	}

	if (operation == OP_MOVE) {
		Window root, win;

		XQueryPointer(d, rootwin, &root, &win, &px, &py, &i, &j,
			&state);
		a = x, b = y;
	} else {
		a = ((ow = hw->child_w) - hw->base_w) / hw->inc_w;
		b = ((oh = hw->child_h) - hw->base_h) / hw->inc_h;
	}

	state = 0;
	XGrabServer(d);
	XMapRaised(d, indicator);

	while (!done) {
		update_indicator(operation, a, b);
		draw_band(x, y, width, height);
		if (operation == OP_MOVE)
			i = px, j = py;
get_input:
		XMaskEvent(d, KeyPressMask | ButtonPressMask, &e);
		if (e.type == ButtonPress)
			break;
		if (e.xkey.state != state) {
			scale = ((state = e.xkey.state) & Mod1Mask) ?
				DEF_SCALE_MOD1 : 0;
			if (state & ControlMask)
				scale += DEF_SCALE_CTRL;
			if (state & ShiftMask)
				scale += DEF_SCALE_SHIFT;
			if (scale == 0)
				scale = DEF_SCALE_NORM;
		}
		switch (XLookupKeysym(&e.xkey, 0)) {
		case XK_Down:
		case XK_j:
		case XK_n:
			dir = 1;
			break;
		case XK_Up:
		case XK_k:
		case XK_p:
			dir = -1;
			break;
		case XK_Left:
		case XK_h:
		case XK_b:
			dir = -1;
			changed_x = True;
			break;
		case XK_Right:
		case XK_l:
		case XK_f:
			dir = 1;
			changed_x = True;
			break;
		case XK_Return:
			done = True;
			continue;
		default:
			goto get_input;
		}
		draw_band(x, y, width, height);
		if (operation == OP_MOVE) {
			if (changed_x == True) {
				changed_x = False;
				if ((px += dir * scale) < 0)
					px = 0;
				else if (px >= scr_w)
					px = scr_w - 1;
				x += px - i;
				a = x;
			} else {
				if ((py += dir * scale) < 0)
					py = 0;
				else if (py >= scr_h)
					py = scr_h - 1;
				y += py - j;
				b = y;
			}
			XWarpPointer(d, None, rootwin, 0, 0, 0, 0, px, py);
		} else if (changed_x == True) {
			changed_x = False;
			i = hw->child_w + (scale < hw->inc_w ? hw->inc_w :
				scale) * dir;
			adjust(hw, &i, NULL, &a, NULL, ADJUST_WIDTH);
			hw->child_w = i;
			width = hw->child_w + 2 *
#ifdef FRAME
				(hw->child_bw + frame_w);
#else
				hw->child_bw;
#endif
		} else {
			i = hw->child_h + (scale < hw->inc_h ? hw->inc_h :
				scale) * dir;
			adjust(hw, NULL, &i, NULL, &b, ADJUST_HEIGHT);
			hw->child_h = i;
			height = hw->child_h + 2 *
#ifdef FRAME
				(hw->child_bw + frame_w) + title_h;
#else
				hw->child_bw;
#endif
		}
	}

	draw_band(x, y, width, height);
	XUnmapWindow(d, indicator);
	if (operation == OP_MOVE) {
		if (child_or_icon == CHILD) {
			hw->x = x;
			hw->y = y;
			move(hw);
		} else {
			XMoveWindow(d, hw->icon, x, y);
			hw->icon_x = x;
			hw->icon_y = y;
		}
	} else if (hw->child_w != ow || hw->child_h != oh) {
#ifdef MAXIMIZE
		hw->flags &= ~MAXIMIZED;
#endif
		XResizeWindow(d, hw->child, hw->child_w, hw->child_h);
#ifdef FRAME
		hw->w = hw->child_w + (hw->child_bw + frame_w) * 2;
		hw->h = title_h + frame_w * 2 + ((hw->flags & SHAPED) ? 0 :
			hw->child_h + hw->child_bw * 2);
		XResizeWindow(d, hw->frame, hw->w, hw->h);
#endif
	}
	XUngrabServer(d);
	XUngrabKeyboard(d, CurrentTime);
ungrab_pointer_and_return:
	XUngrabPointer(d, CurrentTime);
}

void iconify(struct heliwm_window *hw)
{
	if (hw->icon == None) {
		XSetWindowAttributes attr;
		Bool created = False;
		unsigned int w, h;
		unsigned long mask = CWBorderPixel | CWEventMask | CWCursor;

		attr.border_pixel = bdc;
		attr.event_mask = ButtonPressMask | ExposureMask;
		attr.cursor = cursor;
		if (hw->icon_pm != None) {
			Window win;
			int x, y, depth;

			XGetGeometry(d, hw->icon_pm, &win, &x, &y, &w, &h,
				(unsigned int *) &hw->icon_bw, (unsigned int *)
				&depth);
			if (depth != DefaultDepth(d, scr)) {
				attr.background_pixmap = XCreatePixmap(d,
					rootwin, w, h, DefaultDepth(d, scr));
				XCopyPlane(d, hw->icon_pm,
					attr.background_pixmap, text_gc, 0, 0,
					w, h, 0, 0, 1);
				created = True;
			} else
				attr.background_pixmap = hw->icon_pm;
			mask |= CWBackPixmap;
		} else {
			w =
#ifdef NLS
				XmbTextEscapement(fontset, hw->icon_name,
				strlen(hw->icon_name))
#else
				XTextWidth(font_info, hw->icon_name,
				strlen(hw->icon_name))
#endif
				+ H_OFFSET * 2;
			h = title_h;
			attr.background_pixel = bgc;
			mask |= CWBackPixel;
			hw->flags |= NEED_TO_DRAW_ICON;
		}
		hw->icon = XCreateWindow(d, rootwin, hw->x, hw->y, w, h, bdw,
			CopyFromParent, InputOutput, CopyFromParent, mask,
			&attr);
		hw->icon_w = w;
		hw->icon_h = h;
		hw->icon_bw = bdw;
		hw->flags |= CREATED_ICON;
		if (created)
			XFreePixmap(d, attr.background_pixmap);
	}

#ifdef FRAME
	XUnmapWindow(d, hw->frame);
	if (hw->flags & SHAPED) {
		XUnmapWindow(d, hw->child);
		hw->flags |= IGNORE_UNMAP;
	}
#else
	XUnmapWindow(d, hw->child);
	hw->flags |= IGNORE_UNMAP;
#endif
	if (!(hw->flags & ICON_POS_DEFINED)) {
		if (hw->x < 0)
			hw->icon_x = 0;
		else
			hw->icon_x = (hw->x >= scr_w) ? scr_w - 1 : hw->x;
		if (hw->y < 0)
			hw->icon_y = 0;
		else
			hw->icon_y = (hw->y >= scr_h) ? scr_h - 1 : hw->y;
		XWarpPointer(d, None, rootwin, 0, 0, 0, 0, hw->icon_x,
			hw->icon_y);
		if (options & INTERACTIVE_PLACEMENT)
			move_or_resize(hw, OP_MOVE, ICON);
		else
			XMoveWindow(d, hw->icon, hw->icon_x, hw->icon_y);
		hw->flags |= ICON_POS_DEFINED;
	}
	XMapRaised(d, hw->icon);
	hw->flags |= ICONIFIED;
}

#ifdef MAXIMIZE
static void maximize(struct heliwm_window *hw)
{
	int w, h;

	w = scr_w - 2 *
#ifdef FRAME
		(hw->child_bw + frame_w);
#else
		hw->child_bw;
#endif
	h = scr_h - 2 *
#ifdef FRAME
		(hw->child_bw + frame_w) - title_h;
#else
		hw->child_bw;
#endif

	adjust(hw, &w, &h, NULL, NULL, ADJUST_WIDTH | ADJUST_HEIGHT);

	if (w != hw->child_w || h != hw->child_h) {
		hw->flags |= MAXIMIZED;
		hw->ow = hw->child_w;
		hw->oh = hw->child_h;
		hw->ox = hw->x;
		hw->oy = hw->y;
		hw->child_w = w;
		hw->child_h = h;
		if (hw->x != 0 || hw->y != 0) {
			hw->x = hw->y = 0;
			move(hw);
		}
		XResizeWindow(d, hw->child, (unsigned int) w, (unsigned int) h);
#ifdef FRAME
		hw->w = (hw->child_bw + frame_w) * 2 + w;
		hw->h = ((hw->flags & SHAPED) ? 0 : h + hw->child_bw * 2) +
			frame_w * 2 + title_h;
		XResizeWindow(d, hw->frame, hw->w, hw->h);
#endif
	}
}
#endif

void exec_operation(Window win, int op_code)
{
	Window w;
	struct heliwm_window *hw;
	int rel;

	if (op_code != OP_NEXT && op_code != OP_PREV) {
		if (win == None || win == rootwin)
			return;
		rel = find_window(win, &hw);
		if (rel == UNDEFINED)
			return;
	}

	switch (op_code) {
	case OP_MOVE:
		if (rel == ICON)
			move_or_resize(hw, OP_MOVE, ICON);
		else
			move_or_resize(hw, OP_MOVE, CHILD);
		break;
	case OP_RESIZE:
		if (rel == CHILD)
			move_or_resize(hw, OP_RESIZE, CHILD);
		break;
	case OP_DELETE:
		if ((hw->flags & DELETE_WINDOW_IS_SET) && rel == CHILD)
			send_cmessage(hw->child, wm_delete_window);
		break;
	case OP_ICONIFY:
		if (rel == CHILD)
			iconify(hw);
		else if (rel == ICON) {
			XUnmapWindow(d, hw->icon);
#ifdef FRAME
#ifdef SHAPE
			if (hw->flags & SHAPED)
				XMapRaised(d, hw->child);
#endif
			XMapRaised(d, hw->frame);
#else
			XMapRaised(d, hw->child);
#endif
			hw->flags &= ~ICONIFIED;
		}
		if (focus == hw)
			set_focus(focus);
		break;
#ifdef MAXIMIZE
	case OP_MAXIMIZE:
		if (rel != CHILD)
			break;
		XGrabServer(d);
		if (hw->flags & MAXIMIZED) {
			hw->flags &= ~MAXIMIZED;
			if (hw->x != hw->ox || hw->y != hw->oy) {
				hw->x = hw->ox;
				hw->y = hw->oy;
				move(hw);
			}
			if (hw->child_w != hw->ow || hw->child_h != hw->oh) {
				hw->child_w = hw->ow;
				hw->child_h = hw->oh;
				XResizeWindow(d, hw->child, (unsigned int)
					hw->child_w, (unsigned int)
					hw->child_h);
#ifdef FRAME
				hw->w = (hw->child_bw + frame_w) * 2 +
					hw->child_w;
				hw->h = ((hw->flags & SHAPED) ? 0 : hw->child_h
					+ hw->child_bw * 2) + frame_w * 2 +
					title_h;
				XResizeWindow(d, hw->frame, (unsigned int)
					hw->w, (unsigned int) hw->h);
#endif
			}
		} else
			maximize(hw);
		XUngrabServer(d);
		break;
#endif /* MAXIMIZE */
	case OP_NEXT:
	case OP_PREV:
		if (list == NULL)
			break;
		if (op_code == OP_NEXT) {
			if (pointer == NULL || (pointer = pointer->next) ==
				NULL)
				pointer = list;
		} else if (pointer == NULL || (pointer = pointer->prev) == NULL)
			pointer = last;
		hw = pointer;
		if (op_code == OP_NEXT)
			while (!(pointer->flags & MAPPED)) {
				if ((pointer = pointer->next) == NULL)
					pointer = list;
				if (pointer == hw)
					break;
			}
		else
			while (!(pointer->flags & MAPPED)) {
				if ((pointer = pointer->prev) == NULL)
					pointer = last;
				if (pointer == hw)
					break;
			}
		if (pointer->flags & MAPPED) {
			int x, y;

			if (pointer->flags & ICONIFIED) {
				w = pointer->icon;
				x = pointer->icon_x;
				y = pointer->icon_y;
			} else {
#ifdef FRAME
				w = pointer->frame;
#else
				w = pointer->child;
#endif
				x = pointer->x;
				y = pointer->y;
			}
			XWarpPointer(d, None, rootwin, 0, 0, 0, 0, x, y);
			if ((options & FOCUS_WHEN_CIRCULATED) && focus !=
				pointer)
				set_focus(pointer);
			if (options & RAISE_WHEN_CIRCULATED) {
				hw = pointer;
				goto raise_heliwm_window;
			}
		} else {
			pointer = NULL;
			set_focus(NULL);
		}
		break;
	case OP_RAISE:
	case OP_LOWER:
		w = (rel == ICON) ? hw->icon :
#ifdef FRAME
			hw->frame;
#else
			hw->child;
#endif
		if (op_code == OP_RAISE) {
raise_heliwm_window:
			XRaiseWindow(d, w);
#ifdef SHAPE
			if (hw->flags & SHAPED)
				XRaiseWindow(d, hw->child);
#endif
			if ((options & FOCUS_WHEN_RAISED) && focus != hw)
				set_focus(hw);
		} else {
			XLowerWindow(d, w);
#ifdef SHAPE
			if (hw->flags & SHAPED)
				XLowerWindow(d, hw->child);
#endif
		}
		break;
	}
}
